
from siliconcompiler.tools.openroad.openroad import setup as setup_tool
from siliconcompiler.tools.openroad.openroad import build_pex_corners
from siliconcompiler.tools.openroad.openroad import post_process as or_post_process
from siliconcompiler.tools.openroad.openroad import pre_process as or_pre_process
from siliconcompiler.tools._common.asic import set_tool_task_var, get_libraries
from siliconcompiler.tools.openroad.openroad import _set_reports, set_pnr_inputs, set_pnr_outputs
from siliconcompiler.tools._common import get_tool_task


def setup(chip):
    '''
    Generate abstract views (LEF), timing libraries (liberty files),
    circuit descriptions (CDL), and parasitic annotation files (SPEF)
    '''

    # Generic tool setup.
    setup_tool(chip)

    set_pnr_inputs(chip)
    set_pnr_outputs(chip)

    tool = 'openroad'
    design = chip.top()
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    _, task = get_tool_task(chip, step, index)

    # Set thread count to 1 while issue related to write_timing_model segfaulting
    # when multiple threads are on is resolved.
    chip.set('tool', tool, 'task', task, 'threads', 1,
             step=step, index=index, clobber=True)

    stackup = chip.get('option', 'stackup')
    pdk = chip.get('option', 'pdk')

    targetlibs = get_libraries(chip, 'logic')
    macrolibs = get_libraries(chip, 'macro')

    # Determine if exporting the cdl
    set_tool_task_var(chip, param_key='write_cdl',
                      default_value='false',
                      schelp='true/false, when true enables writing the CDL file for the design')
    do_cdl = chip.get('tool', tool, 'task', task, 'var', 'write_cdl',
                      step=step, index=index)[0] == 'true'

    if do_cdl:
        chip.add('tool', tool, 'task', task, 'output', design + '.cdl', step=step, index=index)
        for lib in targetlibs + macrolibs:
            chip.add('tool', tool, 'task', task, 'require',
                     ",".join(['library', lib, 'output', stackup, 'cdl']),
                     step=step, index=index)

    set_tool_task_var(chip, param_key='write_spef',
                      default_value='true',
                      schelp='true/false, when true enables writing the SPEF file for the design')
    do_spef = chip.get('tool', tool, 'task', task, 'var', 'write_spef',
                       step=step, index=index)[0] == 'true'
    set_tool_task_var(chip, param_key='use_spef',
                      default_value=do_spef,
                      schelp='true/false, when true enables reading in SPEF files.')

    if do_spef:
        # Require openrcx pex models
        for corner in chip.get('tool', tool, 'task', task, 'var', 'pex_corners',
                               step=step, index=index):
            chip.add('tool', tool, 'task', task, 'require',
                     ",".join(['pdk', pdk, 'pexmodel', 'openroad-openrcx', stackup, corner]),
                     step=step, index=index)

        # Add outputs SPEF in the format {design}.{pexcorner}.spef
        for corner in chip.get('tool', tool, 'task', task, 'var', 'pex_corners',
                               step=step, index=index):
            chip.add('tool', tool, 'task', task, 'output', design + '.' + corner + '.spef',
                     step=step, index=index)

    # Add outputs LEF
    chip.add('tool', tool, 'task', task, 'output', design + '.lef', step=step, index=index)

    set_tool_task_var(chip, param_key='write_liberty',
                      default_value='true',
                      schelp='true/false, when true enables writing the liberty '
                             'timing model for the design')
    do_liberty = chip.get('tool', tool, 'task', task, 'var', 'write_liberty',
                          step=step, index=index)[0] == 'true'

    if do_liberty:
        # Add outputs liberty model in the format {design}.{libcorner}.lib
        for corner in chip.getkeys('constraint', 'timing'):
            chip.add('tool', tool, 'task', task, 'output', design + '.' + corner + '.lib',
                     step=step, index=index)

    set_tool_task_var(chip, param_key='write_sdf',
                      default_value='true',
                      schelp='true/false, when true enables writing the SDF timing model '
                             'for the design')
    do_sdf = chip.get('tool', tool, 'task', task, 'var', 'write_sdf',
                      step=step, index=index)[0] == 'true'
    if do_sdf:
        # Add outputs liberty model in the format {design}.{libcorner}.sdf
        for corner in chip.getkeys('constraint', 'timing'):
            chip.add('tool', tool, 'task', task, 'output', design + '.' + corner + '.sdf',
                     step=step, index=index)

    _set_reports(chip, [
        'setup',
        'hold',
        'unconstrained',
        'clock_skew',
        'power',
        'drv_violations',
        'fmax',

        # Images
        'placement_density',
        'routing_congestion',
        'power_density',
        'ir_drop',
        'clock_placement',
        'clock_trees',
        'optimization_placement'
    ])


def pre_process(chip):
    or_pre_process(chip)
    build_pex_corners(chip)


def post_process(chip):
    or_post_process(chip)



'''
nextpnr is a vendor neutral FPGA place and route tool with
support for the ICE40, ECP5, and Nexus devices from Lattice.

Documentation: https://github.com/YosysHQ/nextpnr

Sources: https://github.com/YosysHQ/nextpnr

Installation: https://github.com/YosysHQ/nextpnr
'''


#####################################################################
# Make Docs
#####################################################################
def make_docs(chip):
    from siliconcompiler.tools.nextpnr.apr import setup
    setup(chip)
    return chip


################################
#  Custom runtime options
################################
def runtime_options(chip):
    ''' Custom runtime options, returns list of command line options.
    '''
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')

    partname = chip.get('fpga', 'partname')
    topmodule = chip.top()

    options = []

    options.append('--json inputs/' + topmodule + '.netlist.json')
    options.append('--asc outputs/' + topmodule + '.asc')

    if partname == 'ice40up5k-sg48':
        options.append('--up5k --package sg48')

    for constraint_file in chip.find_files('input', 'constraint', 'pcf', step=step, index=index):
        options.append('--pcf ' + constraint_file)

    return options


################################
# Version Check
################################
def parse_version(stdout):
    # Examples:
    # nextpnr-ice40 -- Next Generation Place and Route (Version c73d4cf6)
    # nextpnr-ice40 -- Next Generation Place and Route (Version nextpnr-0.2)
    version = stdout.split()[-1].rstrip(')')
    if version.startswith('nextpnr-'):
        return version.split('-')[1]
    else:
        return version



import os
from siliconcompiler import SiliconCompilerError
from siliconcompiler.tools.vpr import vpr
from siliconcompiler.tools._common import get_tool_task


def setup(chip, clobber=True):
    '''
    Show placed and/or routed designs in VPR GUI
    '''

    tool = 'vpr'
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    _, task = get_tool_task(chip, step, index)

    vpr.setup_tool(chip, clobber=clobber)

    chip.set('tool', tool, 'task', task, 'threads', os.cpu_count(),
             step=step, index=index, clobber=False)


def runtime_options(chip):
    '''Command line options to vpr for show
    '''

    options = generic_show_options(chip)

    options.append('--disp on')

    return options


def generic_show_options(chip):
    ''' Helper function to setup options for show and screenshot
    '''

    design = chip.top()

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    options = vpr.runtime_options(chip)

    file_path = chip.get('tool', tool, 'task', task, 'var', 'show_filepath',
                         step=step, index=index)
    if not file_path:
        raise SiliconCompilerError("Please provide a place or route file", chip=chip)

    if os.path.exists(file_path[0]):
        file_dir = os.path.dirname(file_path[0])
        blif_file = os.path.join(file_dir, f'{design}.blif')
        net_file = os.path.join(file_dir, f'{design}.net')
        place_file = os.path.join(file_dir, f'{design}.place')
        route_file = os.path.join(file_dir, f'{design}.route')
    else:
        raise SiliconCompilerError("Invalid filepath", chip=chip)

    if os.path.exists(blif_file):
        options.append(blif_file)
    else:
        raise SiliconCompilerError("Blif file does not exist", chip=chip)

    if os.path.exists(net_file):
        options.append(f'--net_file {net_file}')
    else:
        raise SiliconCompilerError("Net file does not exist", chip=chip)

    if os.path.exists(route_file) and os.path.exists(place_file):
        options.append('--analysis')
        options.append(f'--place_file {place_file}')
        options.append(f'--route_file {route_file}')
    elif os.path.exists(place_file):
        # NOTE: This is a workaround to display the VPR GUI on the output of the place step.
        # VPR GUI can be invoked during the place, route or analysis steps - not after they are run.
        # Analysis can only be run after route. Hence, the only way to display the output
        # of the is to run the route step. Performing routing could take a significant amount
        # of time, which would not be useful if the user is simply looking to visualize
        # the placed design. Setting max_router_iterations to 0 avoids running routing iterations
        # and provides a fast way to invoke VPR GUI on the placed design.
        options.append('--route')
        options.append('--max_router_iterations 0')
        options.append(f'--place_file {place_file}')
    else:
        raise SiliconCompilerError("Place file does not exist", chip=chip)

    return options



import siliconcompiler
from lambdapdk.interposer.libs.bumps import setup


#########################
if __name__ == "__main__":
    lib = setup(siliconcompiler.Chip('<lib>'))
    lib.write_manifest(f'{lib.top()}.json')




'''
Netgen is a tool for comparing netlists. By comparing a Verilog netlist with
one extracted from a circuit layout, it can be used to perform LVS
verification.

Documentation: http://www.opencircuitdesign.com/netgen/

Installation: https://github.com/RTimothyEdwards/netgen

Sources: https://github.com/RTimothyEdwards/netgen
'''


####################################################################
# Make Docs
####################################################################
def make_docs(chip):
    from siliconcompiler.tools.netgen.lvs import setup
    setup(chip)
    return chip


################################
# Version Check
################################
def parse_version(stdout):
    # First line: Netgen 1.5.190 compiled on Fri Jun 25 16:05:36 EDT 2021
    return stdout.split()[1]


##################################################
if __name__ == "__main__":

    chip = make_docs()
    chip.write_manifest("netgen.json")




from siliconcompiler.tools.klayout.klayout import setup as setup_tool
from siliconcompiler.tools.klayout.screenshot import setup_gui_screenshot
from siliconcompiler.tools._common import input_provides, get_tool_task
from siliconcompiler.tools._common.asic import get_libraries


def setup(chip):
    '''
    Generate a GDSII file from an input DEF file
    '''

    # Generic tool setup.
    setup_tool(chip)

    tool = 'klayout'
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    _, task = get_tool_task(chip, step, index)
    clobber = False

    chip.set('tool', tool, 'task', task, 'threads', 1, step=step, index=index, clobber=clobber)

    script = 'klayout_export.py'
    option = ['-z', '-nc', '-rx', '-r']
    chip.set('tool', tool, 'task', task, 'script', script, step=step, index=index, clobber=clobber)
    chip.set('tool', tool, 'task', task, 'option', option, step=step, index=index, clobber=clobber)

    targetlibs = get_libraries(chip, 'logic')
    stackup = chip.get('option', 'stackup')
    pdk = chip.get('option', 'pdk')

    # Set stream extension
    streams = ('gds', 'oas')
    chip.set('tool', tool, 'task', task, 'var', 'stream', 'gds',
             step=step, index=index, clobber=False)
    chip.set('tool', tool, 'task', task, 'var', 'stream',
             f'Extension to use for stream generation ({streams})',
             field='help')
    default_stream = chip.get('tool', tool, 'task', task, 'var', 'stream',
                              step=step, index=index)[0]
    sc_stream_order = [default_stream, *[s for s in streams if s != default_stream]]

    if stackup:
        macrolibs = get_libraries(chip, 'macro')

        chip.add('tool', tool, 'task', task, 'require', ",".join(['option', 'stackup']),
                 step=step, index=index)
        req_set = False
        for s in sc_stream_order:
            if chip.valid('pdk', pdk, 'layermap', 'klayout', 'def', s, stackup):
                chip.add('tool', tool, 'task', task, 'require',
                         ",".join(['pdk', pdk, 'layermap', 'klayout', 'def', s, stackup]),
                         step=step, index=index)
                req_set = True
                break
        if not req_set:
            # add default require
            chip.add('tool', tool, 'task', task, 'require',
                     ",".join(['pdk', pdk, 'layermap', 'klayout', 'def', 'klayout', stackup]),
                     step=step, index=index)

        for lib in (targetlibs + macrolibs):
            lib_requires_stream = True
            if chip.valid('library', lib, 'option', 'var', 'klayout_allow_missing_cell') and \
               chip.get('library', lib, 'option', 'var', 'klayout_allow_missing_cell'):
                lib_requires_stream = False
            req_set = False
            for s in sc_stream_order:
                if chip.valid('library', lib, 'output', stackup, s):
                    chip.add('tool', tool, 'task', task, 'require',
                             ",".join(['library', lib, 'output', stackup, s]),
                             step=step, index=index)
                    req_set = True
                    break
            if not req_set and lib_requires_stream:
                chip.add('tool', tool, 'task', task, 'require',
                         ",".join(['library', lib, 'output', stackup, default_stream]),
                         step=step, index=index)
            chip.add('tool', tool, 'task', task, 'require',
                     ",".join(['library', lib, 'output', stackup, 'lef']),
                     step=step, index=index)
    else:
        chip.error('Stackup parameter required for Klayout.')

    if not targetlibs:
        chip.add('tool', tool, 'task', task, 'require',
                 'option,var,klayout_libtype',
                 step=step, index=index)

    # Input/Output requirements for default flow
    design = chip.top()
    if design + '.def' in input_provides(chip, step, index):
        chip.add('tool', tool, 'task', task, 'input', design + '.def',
                 step=step, index=index)
    else:
        chip.add('tool', tool, 'task', task, 'require', 'input,layout,def',
                 step=step, index=index)

    chip.add('tool', tool, 'task', task, 'output', f'{design}.{default_stream}',
             step=step, index=index)
    chip.add('tool', tool, 'task', task, 'output', f'{design}.lyt',
             step=step, index=index)
    chip.add('tool', tool, 'task', task, 'output', f'{design}.lyp',
             step=step, index=index)

    # Export GDS with timestamps by default.
    chip.set('tool', tool, 'task', task, 'var', 'timestamps', 'true',
             step=step, index=index, clobber=False)
    chip.set('tool', tool, 'task', task, 'var', 'timestamps',
             'Export GDSII with timestamps',
             field='help')

    chip.set('tool', tool, 'task', task, 'var', 'screenshot', 'true',
             step=step, index=index, clobber=False)
    chip.set('tool', tool, 'task', task, 'var', 'screenshot',
             'true/false: true will cause KLayout to generate a screenshot of the layout',
             field='help')

    if chip.get('tool', tool, 'task', task, 'var', 'screenshot',
                step=step, index=index) == ['true']:
        setup_gui_screenshot(chip, require_input=False)



import xml.etree.ElementTree as ET
import xml.dom.minidom


def generate_vpr_constraints_xml_file(pin_map, filename):

    constraints_xml = generate_vpr_constraints_xml(pin_map)
    write_vpr_constraints_xml_file(constraints_xml, filename)


def generate_vpr_constraints_xml(pin_map):

    constraints_xml = ET.Element("vpr_constraints")

    # Generate partition list section
    partition_list = generate_partition_list_xml(pin_map)

    constraints_xml.append(partition_list)

    return constraints_xml


def generate_partition_list_xml(pin_map):

    partition_list = ET.Element("partition_list")

    # ***ASSUMPTION:  pin map is a dictionary of block names
    #                 and tuples, where the tuples specify the
    #                 (X,Y,subtile) locations that each block
    #                 is constrained to
    for pin, region in pin_map.items():

        cur_partition = generate_partition_xml(pin, region)
        partition_list.append(cur_partition)

    return partition_list


def generate_partition_xml(pin, pin_region):

    partition = ET.Element("partition")

    partition_name = generate_partition_name(pin)
    partition.set("name", partition_name)

    atom_xml = generate_add_atom_xml(pin)
    partition.append(atom_xml)

    x_low, x_high, y_low, y_high, subtile = generate_region_from_pin(pin_region)
    region_xml = generate_add_region_xml(x_low, x_high, y_low, y_high, subtile)
    partition.append(region_xml)

    return partition


def generate_region_from_pin(pin_region):

    # ***ASSUMPTION:  Pin region is a 3-element tuple
    #                 containing (X,Y,subtile) coordinates

    # TODO figure out a scheme that supports VPR's notion
    # of specifying a region size of > 1x1
    x_low = int(pin_region[0])
    x_high = int(pin_region[0])
    y_low = int(pin_region[1])
    y_high = int(pin_region[1])
    subtile = int(pin_region[2])
    return x_low, x_high, y_low, y_high, subtile


def generate_partition_name(pin):

    partition_name = pin
    partition_name = partition_name.replace('[', '_')
    partition_name = partition_name.replace(']', '')
    partition_name = "part_" + partition_name

    return partition_name


def generate_add_atom_xml(pin_name):

    atom_xml = ET.Element("add_atom")

    atom_xml.set("name_pattern", str(pin_name))

    return atom_xml


def generate_add_region_xml(x_low, x_high, y_low, y_high, subtile):

    region_xml = ET.Element("add_region")

    region_xml.set("x_low", str(x_low))
    region_xml.set("y_low", str(y_low))
    region_xml.set("x_high", str(x_high))
    region_xml.set("y_high", str(y_high))
    region_xml.set("subtile", str(subtile))

    return region_xml


def write_vpr_constraints_xml_file(constraints: ET.Element, filename: str):

    dom = xml.dom.minidom.parseString(ET.tostring(constraints))
    xml_string = dom.toprettyxml()

    with open(filename, 'w') as xfile:
        xfile.write(str(xml_string))



import os
import shutil
import glob
from siliconcompiler.tools._common import add_frontend_requires, get_tool_task, has_input_files
from siliconcompiler import sc_open, SiliconCompilerError


def setup(chip):
    '''
    Performs high level synthesis to generate a verilog output
    '''

    if not has_input_files(chip, 'input', 'config', 'chisel') and \
       not has_input_files(chip, 'input', 'hll', 'scala'):
        return "no files in [input,hll,scala] or [input,config,chisel]"

    tool = 'chisel'
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    _, task = get_tool_task(chip, step, index)

    # Standard Setup
    refdir = 'tools/' + tool
    chip.set('tool', tool, 'exe', 'sbt')
    chip.set('tool', tool, 'vswitch', '--version')
    chip.set('tool', tool, 'version', '>=1.5.5', clobber=False)

    chip.set('tool', tool, 'task', task, 'refdir', refdir,
             step=step, index=index,
             package='siliconcompiler', clobber=False)
    chip.set('tool', tool, 'task', task, 'threads', os.cpu_count(),
             step=step, index=index, clobber=False)

    chip.set('tool', tool, 'task', task, 'option', ['-batch',
                                                    '--no-share',
                                                    '--no-global'],
             step=step, index=index)

    chip.set('tool', tool, 'task', task, 'var', 'application',
             'Application name of the chisel program',
             field='help')
    chip.set('tool', tool, 'task', task, 'var', 'argument',
             'Arguments for the chisel build',
             field='help')

    # Input/Output requirements
    if chip.valid('input', 'config', 'chisel') and \
       chip.get('input', 'config', 'chisel', step=step, index=index):
        chip.add('tool', tool, 'task', task, 'require', 'input,config,chisel',
                 step=step, index=index)
        if len(chip.get('input', 'config', 'chisel', step=step, index=index)) != 1:
            raise SiliconCompilerError('Only one build.sbt is supported.', chip=chip)

    if chip.valid('input', 'hll', 'scala') and \
       chip.get('input', 'hll', 'scala', step=step, index=index):
        chip.add('tool', tool, 'task', task, 'require', 'input,hll,scala',
                 step=step, index=index)

    chip.add('tool', tool, 'task', task, 'output', chip.top() + '.v', step=step, index=index)

    add_frontend_requires(chip, [])


def pre_process(chip):
    tool = 'chisel'
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    _, task = get_tool_task(chip, step, index)

    refdir = chip.find_files('tool', tool, 'task', task, 'refdir', step=step, index=index)[0]

    if chip.valid('input', 'config', 'chisel') and \
       chip.get('input', 'config', 'chisel', step=step, index=index):
        build_file = chip.find_files('input', 'config', 'chisel', step=step, index=index)[0]
        work_dir = chip.getworkdir(step=step, index=index)
        build_dir = os.path.dirname(build_file)
        # Expect file tree from: https://www.scala-sbt.org/1.x/docs/Directories.html
        # copy build.sbt
        # copy src/
        shutil.copyfile(build_file, os.path.join(work_dir, 'build.sbt'))
        shutil.copytree(os.path.join(build_dir, 'src'),
                        os.path.join(work_dir, 'src'))
        if os.path.exists(os.path.join(build_dir, 'project')):
            shutil.copytree(os.path.join(build_dir, 'project'),
                            os.path.join(work_dir, 'project'))
        return

    for filename in ('build.sbt', 'SCDriver.scala'):
        src = os.path.join(refdir, filename)
        dst = filename
        shutil.copyfile(src, dst)

    # Chisel driver relies on Scala files being collected into '$CWD/inputs'
    chip.set('input', 'hll', 'scala', True, field='copy')
    chip.collect(directory=os.path.join(chip.getworkdir(step=step, index=index), 'inputs'))


def runtime_options(chip):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    design = chip.top()

    runMain = ["runMain"]
    if chip.valid('input', 'config', 'chisel') and \
       chip.get('input', 'config', 'chisel', step=step, index=index):
        app = design
        if chip.valid('tool', tool, 'task', task, 'var', 'application') and \
           chip.get('tool', tool, 'task', task, 'var', 'application', step=step, index=index):
            app = chip.get('tool', tool, 'task', task, 'var', 'application',
                           step=step, index=index)[0]

        runMain.append(f"{app}")

        if chip.valid('tool', tool, 'task', task, 'var', 'argument') and \
           chip.get('tool', tool, 'task', task, 'var', 'argument', step=step, index=index):
            runMain.extend(chip.get('tool', tool, 'task', task, 'var', 'argument',
                                    step=step, index=index))
        runMain.append("--")

        runMain.append("--target-dir chisel-output")
    else:
        # Use built in driver
        runMain.append("SCDriver")
        runMain.append(f"--module {chip.top(step=step, index=index)}")

        runMain.append(f"--output-file ../outputs/{design}.v")

    return [f'"{" ".join(runMain)}"']


def post_process(chip):
    chisel_path = 'chisel-output'
    if os.path.exists(chisel_path):
        design = chip.top()
        with open(f'outputs/{design}.v', 'w') as out:
            for f in glob.glob(os.path.join(chisel_path, '*.v')):
                with sc_open(f) as i_file:
                    out.writelines(i_file.readlines())



from siliconcompiler.tools.verilator.verilator import setup as setup_tool
from siliconcompiler.tools.verilator.verilator import runtime_options as runtime_options_tool
from siliconcompiler.tools._common import get_tool_task


def setup(chip):
    '''
    Lints Verilog source. Results of linting can be programmatically queried
    using errors/warnings metrics.
    '''

    # Generic tool setup.
    setup_tool(chip)

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    chip.set('tool', tool, 'task', task, 'output', f'{chip.top()}.v',
             step=step, index=index)


def runtime_options(chip):
    cmdlist = runtime_options_tool(chip)
    cmdlist.append('--xml-only')
    cmdlist.append('--no-timing')
    return cmdlist



'''
Xyce is a high performance SPICE-compatible circuit simulator
capable capable of solving extremely large circuit problems by
supporting large-scale parallel computing platforms. It also
supports serial execution on all common desktop platforms,
and small-scale parallel runs on Unix-like systems.

Documentation: https://xyce.sandia.gov/documentation-tutorials/

Sources: https://xyce.sandia.gov/downloads/source-code/

Installation: https://xyce.sandia.gov/documentation-tutorials/building-guide/

Status: SC integration WIP
'''
from siliconcompiler.tools._common import get_tool_task


################################
# Setup Tool (pre executable)
################################
def setup(chip):

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    chip.set('tool', tool, 'exe', 'Xyce')
    chip.set('tool', tool, 'vswitch', '-v')
    chip.set('tool', tool, 'version', '>=v7.8')

    chip.set('tool', tool, 'task', task, 'regex', 'warnings', 'warning',
             step=step, index=index, clobber=False)
    chip.set('tool', tool, 'task', task, 'regex', 'errors', 'error',
             step=step, index=index, clobber=False)


def parse_version(stdout):
    ver = stdout.split()[-1]
    return ver.split('-')[0]



from siliconcompiler.tools._common import input_provides, add_common_file, get_tool_task
from siliconcompiler.tools._common.asic import set_tool_task_var
from siliconcompiler.tools.openroad.openroad import setup as setup_tool
from siliconcompiler.tools.openroad.openroad import build_pex_corners
from siliconcompiler.tools.openroad.openroad import post_process as or_post_process
from siliconcompiler.tools.openroad.openroad import pre_process as or_pre_process
from siliconcompiler.tools.openroad.openroad import _set_reports, set_pnr_inputs, set_pnr_outputs


def setup(chip):
    '''
    Perform floorplanning, pin placements, macro placements and power grid generation
    '''

    # Generic tool setup.
    setup_tool(chip)

    tool = 'openroad'
    design = chip.top()
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    _, task = get_tool_task(chip, step, index)

    if chip.valid('input', 'asic', 'floorplan') and \
       chip.get('input', 'asic', 'floorplan', step=step, index=index):
        chip.add('tool', tool, 'task', task, 'require',
                 ",".join(['input', 'asic', 'floorplan']),
                 step=step, index=index)

    if f'{design}.vg' in input_provides(chip, step, index):
        chip.add('tool', tool, 'task', task, 'input', design + '.vg',
                 step=step, index=index)
    else:
        chip.add('tool', tool, 'task', task, 'require', 'input,netlist,verilog',
                 step=step, index=index)

    set_pnr_inputs(chip)
    set_pnr_outputs(chip)

    if chip.valid('tool', tool, 'task', task, 'file', 'padring') and \
       chip.get('tool', tool, 'task', task, 'file', 'padring',
                step=step, index=index):
        chip.add('tool', tool, 'task', task, 'require',
                 ','.join(['tool', tool, 'task', task, 'file', 'padring']),
                 step=step, index=index)
    chip.set('tool', tool, 'task', task, 'file', 'padring',
             'script to insert the padring',
             field='help')

    set_tool_task_var(chip, param_key='remove_synth_buffers',
                      default_value=True,
                      schelp='remove buffers inserted by synthesis')

    snap = chip.get('tool', tool, 'task', task, 'var', 'ifp_snap_strategy',
                    step=step, index=index)[0]
    snaps_allowed = ('none', 'site', 'manufacturing_grid')
    if snap not in snaps_allowed:
        chip.error(f'{snap} is not a supported snapping strategy. Allowed values: {snaps_allowed}')

    add_common_file(chip, 'sc_pin_constraint', 'tcl/sc_pin_constraints.tcl')

    _set_reports(chip, [
        'setup',
        'unconstrained',
        'power'
    ])


def pre_process(chip):
    or_pre_process(chip)
    build_pex_corners(chip)


def post_process(chip):
    or_post_process(chip)



import docker
import os
from siliconcompiler.package import get_cache_path
from siliconcompiler.package import _path as sc_path
from siliconcompiler.utils import default_email_credentials_file
from pathlib import Path
import sys


def get_image(chip, step, index):
    from siliconcompiler import __version__

    queue = chip.get('option', 'scheduler', 'queue', step=step, index=index)
    if queue:
        return queue

    return os.getenv(
        'SC_DOCKER_IMAGE',
        f'ghcr.io/siliconcompiler/sc_runner:v{__version__}')


def get_volumes_directories(chip, cache_dir, workdir, step, index):
    all_dirs = set()
    # Collect files
    for key in chip.allkeys():
        sc_type = chip.get(*key, field='type')

        if 'file' in sc_type or 'dir' in sc_type:
            cstep = step
            cindex = index

            if 'never' in chip.get(*key, field='pernode'):
                cstep = None
                cindex = None

            files = chip.find_files(*key, step=cstep, index=cindex, missing_ok=True)
            if files:
                if not isinstance(files, list):
                    files = [files]
                for path in files:
                    if path is None:
                        continue
                    if 'file' in sc_type:
                        all_dirs.add(os.path.dirname(path))
                    else:
                        all_dirs.add(path)

    # Collect caches
    for package in chip.getkeys('package', 'source'):
        all_dirs.add(sc_path(chip, package, None))

    all_dirs = [
        Path(cache_dir),
        Path(workdir),
        Path(chip.scroot),
        *[Path(path) for path in all_dirs]]

    pruned_dirs = all_dirs.copy()
    for base_path in all_dirs:
        if base_path not in pruned_dirs:
            continue

        new_pruned_dirs = [base_path]
        for check_path in pruned_dirs:
            if base_path == check_path:
                continue

            if base_path not in check_path.parents:
                new_pruned_dirs.append(check_path)
        pruned_dirs = new_pruned_dirs

    pruned_dirs = set(pruned_dirs)

    builddir = chip.find_files('option', 'builddir')

    rw_volumes = set()

    for path in pruned_dirs:
        for rw_allow in (Path(builddir), Path(workdir), Path(cache_dir)):
            if path == rw_allow or path in rw_allow.parents:
                rw_volumes.add(path)

    ro_volumes = pruned_dirs.difference(rw_volumes)

    return rw_volumes, ro_volumes


def init(chip):
    if sys.platform == 'win32':
        # this avoids the issue of different file system types
        chip.logger.error('Setting copy field to true for docker run on Windows')
        for key in chip.allkeys():
            if key[0] == 'history':
                continue
            sc_type = chip.get(*key, field='type')
            if 'dir' in sc_type or 'file' in sc_type:
                chip.set(*key, True, field='copy')
        chip.collect()


def run(chip, step, index, replay):
    # Import here to avoid circular import
    from siliconcompiler.scheduler import _haltstep

    try:
        client = docker.from_env()
        client.version()
    except (docker.errors.DockerException, docker.errors.APIError) as e:
        chip.logger.error(f'Unable to connect to docker: {e}')
        _haltstep(chip, chip.get('option', 'flow'), step, index)

    is_windows = sys.platform == 'win32'

    workdir = chip.getworkdir()
    start_cwd = os.getcwd()

    # Remove handlers from logger
    chip.logger.handlers.clear()

    # Reinit logger
    chip._init_logger(step=step, index=index, in_run=True)

    # Change working directory since the run may delete this folder
    os.chdir(workdir)

    image_name = get_image(chip, step, index)

    # Pull image if needed
    try:
        image = client.images.get(image_name)
    except docker.errors.ImageNotFound:
        # Needs a lock to avoid downloading a bunch in parallel
        image_repo, image_tag = image_name.split(':')
        chip.logger.info(f'Pulling docker image {image_name}')
        try:
            image = client.images.pull(image_repo, tag=image_tag)
        except docker.errors.APIError as e:
            chip.logger.error(f'Unable to pull image: {e}')
            image_src = image_repo.split('/')[0]
            chip.logger.error(f"  if you are logged into {image_src} with expired credentials, "
                              f"please use 'docker logout {image_src}'")
            _haltstep(chip, chip.get('option', 'flow'), step, index)

    email_file = default_email_credentials_file()
    if is_windows:
        # Hack to get around manifest merging
        chip.set('option', 'cachedir', None)
        cache_dir = '/sc_cache'
        cwd = '/sc_docker'
        builddir = f'{cwd}/build'

        local_cfg = os.path.join(start_cwd, 'sc_docker.json')
        job = chip.get('option', 'jobname')
        cfg = f'{builddir}/{chip.design}/{job}/{step}/{index}/sc_docker.json'

        user = None

        volumes = [
            f"{chip.cwd}:{cwd}:rw",
            f"{get_cache_path(chip)}:{cache_dir}:rw"
        ]
        chip.logger.debug(f'Volumes: {volumes}')

        env = {}

        if os.path.exists(email_file):
            env["HOME"] = "/sc_home"

            volumes.append(f'{os.path.dirname(email_file)}:/sc_home/.sc:ro')
    else:
        cache_dir = get_cache_path(chip)
        cwd = chip.cwd
        builddir = chip.find_files('option', 'builddir')

        local_cfg = os.path.abspath('sc_docker.json')
        cfg = local_cfg

        user = os.getuid()

        rw_volumes, ro_volumes = get_volumes_directories(chip, cache_dir, workdir, step, index)
        volumes = [
            *[
                f'{path}:{path}:rw' for path in rw_volumes
            ],
            *[
                f'{path}:{path}:ro' for path in ro_volumes
            ]
        ]
        chip.logger.debug(f'Read write volumes: {rw_volumes}')
        chip.logger.debug(f'Read only volumes: {ro_volumes}')

        env = {}
        if os.path.exists(email_file):
            env["HOME"] = "/sc_home"

            volumes.append(f'{os.path.dirname(email_file)}:/sc_home/.sc:ro')

    container = None
    try:
        container = client.containers.run(
            image.id,
            volumes=volumes,
            labels=[
                "siliconcompiler",
                f"sc_node:{chip.design}:{step}{index}"
            ],
            user=user,
            detach=True,
            tty=True,
            auto_remove=True,
            environment=env)

        # Write manifest to make it available to the docker runner
        chip.write_manifest(local_cfg)

        cachemap = []
        for package in chip.getkeys('package', 'source'):
            cachemap.append(f'{package}:{sc_path(chip, package, None)}')

        chip.logger.info(f'Running in docker container: {container.name} ({container.short_id})')
        args = [
            '-cfg', cfg,
            '-cwd', cwd,
            '-builddir', builddir,
            '-cachedir', cache_dir,
            '-step', step,
            '-index', index,
            '-unset_scheduler'
        ]
        if not is_windows and cachemap:
            args.append('-cachemap')
            args.append(' '.join(cachemap))
        cmd = f'python3 -m siliconcompiler.scheduler.run_node {" ".join(args)}'
        exec_handle = client.api.exec_create(container.name, cmd)
        stream = client.api.exec_start(exec_handle, stream=True)

        # Print the log
        for chunk in stream:
            for line in chunk.decode().splitlines():
                print(line)

        if client.api.exec_inspect(exec_handle['Id']).get('ExitCode') != 0:
            _haltstep(chip, chip.get('option', 'flow'), step, index, log=False)
    finally:
        # Ensure we clean up containers
        if container:
            try:
                container.stop()
            except docker.errors.APIError:
                chip.logger.error('Failed to stop docker container')

    # Restore working directory
    os.chdir(start_cwd)



from siliconcompiler.tools.openroad import openroad
from siliconcompiler.tools.openroad.openroad import setup as setup_tool
from siliconcompiler.tools.openroad.openroad import build_pex_corners
from siliconcompiler.tools.openroad.show import copy_show_files, generic_show_setup
from siliconcompiler.tools.openroad.openroad import pre_process as or_pre_process
from siliconcompiler.tools.openroad.openroad import _set_reports
from siliconcompiler.tools._common import get_tool_task


####################################################################
# Make Docs
####################################################################
def make_docs(chip):
    openroad.make_docs(chip)
    chip.set('tool', 'openroad', 'task', 'screenshot', 'var', 'show_filepath', '<path>')


def setup(chip):
    '''
    Generate a PNG file from a layout file
    '''

    tool = 'openroad'
    design = chip.top()
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    _, task = get_tool_task(chip, step, index)

    # Generic tool setup.
    setup_tool(chip)

    generic_show_setup(chip, task, True)

    chip.add('tool', tool, 'task', task, 'output', design + '.png', step=step, index=index)

    chip.set('tool', tool, 'task', task, 'var', 'show_vertical_resolution', '1024',
             step=step, index=index, clobber=False)

    chip.set('tool', tool, 'task', task, 'var', 'include_report_images', 'false',
             step=step, index=index, clobber=False)
    chip.set('tool', tool, 'task', task, 'var', 'include_report_images',
             'true/false, include the images in reports/',
             field='help')

    _set_reports(chip, [
        # Images
        'placement_density',
        'routing_congestion',
        'power_density',
        'ir_drop',
        'clock_placement',
        'clock_trees',
        'optimization_placement'
    ])


def pre_process(chip):
    or_pre_process(chip)
    copy_show_files(chip)
    build_pex_corners(chip)



import os
import re
from siliconcompiler.tools._common import \
    add_require_input, get_input_files, add_frontend_requires, get_frontend_options, \
    get_tool_task, has_input_files
from siliconcompiler.tools.surelog import setup as setup_tool
from siliconcompiler.tools.surelog import runtime_options as runtime_options_tool
from siliconcompiler import sc_open
from siliconcompiler import utils


##################################################
def setup(chip):
    '''
    Import verilog files
    '''

    if not has_input_files(chip, 'input', 'rtl', 'verilog') and \
       not has_input_files(chip, 'input', 'rtl', 'systemverilog'):
        return "no files in [input,rtl,systemverilog] or [input,rtl,verilog]"

    # Generic tool setup.
    setup_tool(chip)

    tool = 'surelog'
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    _, task = get_tool_task(chip, step, index)

    # Runtime parameters.
    chip.set('tool', tool, 'task', task, 'threads', os.cpu_count(),
             step=step, index=index, clobber=False)

    # Input/Output requirements
    chip.set('tool', tool, 'task', task, 'output', __outputfile(chip), step=step, index=index)

    # Schema requirements
    add_require_input(chip, 'input', 'rtl', 'verilog')
    add_require_input(chip, 'input', 'rtl', 'systemverilog')
    add_require_input(chip, 'input', 'cmdfile', 'f')
    add_frontend_requires(chip, ['ydir', 'idir', 'vlib', 'libext', 'define', 'param'])


################################
#  Custom runtime options
################################
def runtime_options(chip):

    ''' Custom runtime options, returnst list of command line options.
    '''

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')

    opts = get_frontend_options(chip,
                                ['ydir',
                                 'idir',
                                 'vlib',
                                 'libext',
                                 'define',
                                 'param'])

    # Command-line options.
    cmdlist = runtime_options_tool(chip)

    # -parse is slow but ensures the SV code is valid
    # we might want an option to control when to enable this
    # or replace surelog with a SV linter for the validate step
    cmdlist.append('-parse')
    # We don't use UHDM currently, so disable. For large designs, this file is
    # very big and takes a while to write out.
    cmdlist.append('-nouhdm')

    libext = opts['libext']
    if libext:
        libext_option = f"+libext+.{'+.'.join(libext)}"
    else:
        # default value for backwards compatibility
        libext_option = '+libext+.sv+.v'
    cmdlist.append(libext_option)

    #####################
    # Library directories
    #####################
    for value in opts['ydir']:
        cmdlist.append('-y ' + value)

    #####################
    # Library files
    #####################
    for value in opts['vlib']:
        cmdlist.append('-v ' + value)

    #####################
    # Include paths
    #####################
    for value in opts['idir']:
        cmdlist.append('-I' + value)

    #######################
    # Variable Definitions
    #######################
    for value in opts['define']:
        cmdlist.append('-D' + value)

    #######################
    # Command files
    #######################
    for value in get_input_files(chip, 'input', 'cmdfile', 'f'):
        cmdlist.append('-f ' + value)

    #######################
    # Sources
    #######################
    for value in get_input_files(chip, 'input', 'rtl', 'systemverilog'):
        cmdlist.append(value)
    for value in get_input_files(chip, 'input', 'rtl', 'verilog'):
        cmdlist.append(value)

    #######################
    # Top Module
    #######################
    cmdlist.append(f'-top {chip.top(step, index)}')

    ###############################
    # Parameters (top module only)
    ###############################
    # Set up user-provided parameters to ensure we elaborate the correct modules
    for param, value in opts['param']:
        cmdlist.append(f'-P{param}={value}')

    return cmdlist


##################################################
def post_process(chip):
    ''' Tool specific function to run after step execution
    '''

    filemap = []
    with sc_open('slpp_all/file_map.lst') as filelist:
        for mapping in filelist:
            filemap.append(mapping)

    def lookup_sources(file):
        for fmap in filemap:
            if fmap.startswith(file):
                return fmap[len(file):].strip()
        return "unknown"

    # https://github.com/chipsalliance/Surelog/issues/3776#issuecomment-1652465581
    surelog_escape = re.compile(r"#~@([a-zA-Z_0-9.\$/\:\[\] ]*)#~@")

    # Look in slpp_all/file_elab.lst for list of Verilog files included in
    # design, read these and concatenate them into one pickled output file.
    output_template = utils.get_file_template('output.v',
                                              root=os.path.join(os.path.dirname(__file__),
                                                                'templates'))

    with sc_open('slpp_all/file_elab.lst') as filelist, \
            open(f'outputs/{__outputfile(chip)}', 'w') as outfile:
        for path in filelist.read().split('\n'):
            path = path.strip('"')
            if not path:
                # skip empty lines
                continue
            with sc_open(path) as infile:
                source_files = lookup_sources(path)
                unescaped_data = surelog_escape.sub(r"\\\1 ", infile.read())

                outfile.write(output_template.render(
                    source_file=source_files,
                    content=unescaped_data
                ))

                outfile.write('\n')


def __outputfile(chip):
    is_systemverilog = has_input_files(chip, 'input', 'rtl', 'systemverilog')
    if is_systemverilog:
        return f'{chip.top()}.sv'
    return f'{chip.top()}.v'



import pandas
import shutil

from siliconcompiler.report.utils import _collect_data, _get_flowgraph_path
from siliconcompiler.tools._common import get_tool_task
from siliconcompiler.utils import truncate_text


def _show_summary_table(chip, flow, flowgraph_nodes, show_all_indices):
    '''
    Prints the end of run summary table
    '''

    # Display data
    column_width = 15

    max_line_width = max(2 * column_width, int(0.95*shutil.get_terminal_size().columns))

    nodes, _, metrics, metrics_unit, metrics_to_show, _ = \
        _collect_data(chip, flow, flowgraph_nodes)

    selected_tasks = \
        _get_flowgraph_path(chip, flow, flowgraph_nodes, only_include_successful=True)

    # only report tool based steps functions
    for (step, index) in list(flowgraph_nodes):
        if get_tool_task(chip, step, index, flow=flow)[0] == 'builtin':
            del flowgraph_nodes[flowgraph_nodes.index((step, index))]

    if show_all_indices:
        nodes_to_show = nodes
    else:
        nodes_to_show = [n for n in nodes if n in selected_tasks]

    # Custom reporting modes
    paramlist = []
    for item in chip.getkeys('option', 'param'):
        paramlist.append(item + "=" + chip.get('option', 'param', item))

    if paramlist:
        paramstr = ', '.join(paramlist)
    else:
        paramstr = "None"

    # trim labels to column width
    column_labels = []
    labels = [f'{step}{index}' for step, index in nodes_to_show]
    if labels:
        column_width = min([column_width, max([len(label) for label in labels])])

    for label in labels:
        column_labels.append(truncate_text(label, column_width).center(column_width))

    row_labels = []
    if metrics_to_show:
        row_label_len = max([len(metric) for metric in metrics_to_show])
        row_unit_len = max([len(metrics_unit[metric]) for metric in metrics_to_show])
        for metric in metrics_to_show:
            row_labels.append(f'{metric:<{row_label_len}}  {metrics_unit[metric]:>{row_unit_len}}')

    data = []
    for metric in metrics_to_show:
        row = []
        for node in nodes_to_show:
            value = metrics[node][metric]
            if value is None:
                value = '---'
            value = ' ' + value.center(column_width)
            row.append(value)
        data.append(row)

    info_list = ["SUMMARY:\n",
                 "design : " + chip.design,
                 "params : " + paramstr,
                 "jobdir : " + chip.getworkdir()]

    pdk = chip.get('option', 'pdk')
    if pdk:
        info_list.extend([
            f"foundry : {chip.get('pdk', pdk, 'foundry')}",
            f"process : {pdk}"])
    else:
        fpga_partname = chip.get('fpga', 'partname')
        if fpga_partname:
            info_list.append(f"partname : {fpga_partname}")

    libraries = set()
    for val, step, index in chip.schema._getvals('asic', 'logiclib'):
        if not step or (step, index) in flowgraph_nodes:
            libraries.update(val)
    if libraries:
        info_list.append(f"targetlibs : {' '.join(libraries)}")

    info = '\n'.join(info_list)

    print("-" * max_line_width)
    print(info, "\n")

    df = pandas.DataFrame(data, row_labels, column_labels)
    if not df.empty:
        print(df.to_string(line_width=max_line_width, col_space=2))
    else:
        print(' No metrics to display!')
    print("-" * max_line_width)



'''
The primary objective of the PandA project is to develop a usable framework that will
enable the research of new ideas in the HW-SW Co-Design field.

The PandA framework includes methodologies supporting the research on high-level synthesis
of hardware accelerators, on parallelism extraction for embedded systems, on hardware/software
partitioning and mapping, on metrics for performance estimation of embedded software
applications and on dynamic reconfigurable devices.

Documentation: https://github.com/ferrandi/PandA-bambu

Sources: https://github.com/ferrandi/PandA-bambu

Installation: https://panda.dei.polimi.it/?page_id=88
'''

from siliconcompiler.tools.bambu import convert


####################################################################
# Make Docs
####################################################################
def make_docs(chip):
    convert.setup(chip)
    return chip


def parse_version(stdout):
    # Long multiline output, but second-to-last line looks like:
    # Version: PandA 0.9.6 - Revision 5e5e306b86383a7d85274d64977a3d71fdcff4fe-main
    version_line = stdout.split('\n')[-3]
    return version_line.split()[2]



import pya
import os
import sys


def show(schema, tech, input_path, output_path, screenshot=False):
    # Extract info from manifest
    flow = schema.get('option', 'flow')
    step = schema.get('arg', 'step')
    index = schema.get('arg', 'index')
    task = schema.get('flowgraph', flow, step, index, 'task')

    if 'hide_layers' in schema.getkeys('tool', 'klayout', 'task', task, 'var'):
        sc_hide_layers = schema.get('tool', 'klayout', 'task', task, 'var', 'hide_layers',
                                    step=step, index=index)
    else:
        sc_hide_layers = []

    # Load KLayout technology file
    layoutOptions = tech.load_layout_options

    # These may be disabled in our KLayout tech file for reasons relating to GDS
    # export, but for the purposes of viewing we'll hardcode them to True.
    layoutOptions.lefdef_config.produce_blockages = True
    layoutOptions.lefdef_config.produce_cell_outlines = True
    layoutOptions.lefdef_config.produce_obstructions = True

    # Always use LEF geometry even when LEF file contains FOREIGN statement.
    layoutOptions.lefdef_config.macro_resolution_mode = 1

    tech.load_layout_options = layoutOptions

    app = pya.Application.instance()
    main_window = pya.MainWindow.instance()
    if not main_window:
        print('[WARNING] unable to show layout as the main window is not available')
        return

    # Opinionated default KLayout configuration
    # see ~/.klayout/klayoutrc for a list of configuration keys

    # show all cells
    app.set_config('full-hierarchy-new-cell', 'true')
    # no tip pop-ups
    app.set_config('tip-window-hidden', ','.join(['only-top-level-shown-by-default=3',
                                                  'editor-mode=4',
                                                  'editor-mode=0']))
    # hide text
    app.set_config('text-visible', 'false')
    # dark background
    app.set_config('background-color', '#212121')

    # Display the file!
    print(f"[INFO] Opening {input_path}")
    cell_view = main_window.load_layout(input_path, tech.name)
    layout_view = cell_view.view()

    # Hide layers that shouldn't be shown in the current view.
    for layer in layout_view.each_layer():
        layer_ldt = f'{layer.source_layer}/{layer.source_datatype}'
        layer_name = layer.source_name
        if not layer_name:
            layer_name = layer.name
        print_layer = layer_name
        if ' ' in layer_name:
            layer_name = layer_name.split()[0]
        if (layer_name in sc_hide_layers) or \
           (layer_ldt in sc_hide_layers):
            print(f"[INFO] Turning off layer: {print_layer} : {layer_ldt}")
            layer.visible = False

    # If 'screenshot' mode is set, save image and exit.
    if screenshot:
        xbins = int(schema.get('tool', 'klayout', 'task', task, 'var', 'xbins',
                               step=step, index=index)[0])
        ybins = int(schema.get('tool', 'klayout', 'task', task, 'var', 'ybins',
                               step=step, index=index)[0])

        if xbins == 1 and ybins == 1:
            __screenshot(schema, layout_view, output_path)
        else:
            __screenshot_montage(schema, layout_view, xbins, ybins)


def __screenshot(schema, layout_view, output_path):
    flow = schema.get('option', 'flow')
    step = schema.get('arg', 'step')
    index = schema.get('arg', 'index')
    task = schema.get('flowgraph', flow, step, index, 'task')

    # Save a screenshot. TODO: Get aspect ratio from sc_cfg?
    horizontal_resolution = int(schema.get('tool', 'klayout', 'task', task, 'var',
                                           'show_horizontal_resolution',
                                           step=step, index=index)[0])
    vertical_resolution = int(schema.get('tool', 'klayout', 'task', task, 'var',
                                         'show_vertical_resolution',
                                         step=step, index=index)[0])

    print(f'[INFO] Saving screenshot to {output_path}')
    layout_view.save_image(output_path, horizontal_resolution, vertical_resolution)


def __screenshot_montage(schema, view, xbins, ybins):
    flow = schema.get('option', 'flow')
    step = schema.get('arg', 'step')
    index = schema.get('arg', 'index')
    task = schema.get('flowgraph', flow, step, index, 'task')

    app = pya.Application.instance()
    options = (
        "grid-show-ruler",
        "grid-visible",
        "text-visible"
    )
    for option in options:
        app.set_config(option, "false")

    app.set_config("background-color", "#000000")  # Black

    design = schema.get('option', 'entrypoint')
    if not design:
        design = schema.get('design')

    horizontal_resolution = int(schema.get('tool', 'klayout', 'task', task, 'var',
                                           'show_horizontal_resolution',
                                           step=step, index=index)[0])
    vertical_resolution = int(schema.get('tool', 'klayout', 'task', task, 'var',
                                         'show_vertical_resolution',
                                         step=step, index=index)[0])
    margin = float(schema.get('tool', 'klayout', 'task', task, 'var',
                              'margin',
                              step=step, index=index)[0])
    linewidth = int(schema.get('tool', 'klayout', 'task', task, 'var',
                               'linewidth',
                               step=step, index=index)[0])
    oversampling = int(schema.get('tool', 'klayout', 'task', task, 'var',
                                  'oversampling',
                                  step=step, index=index)[0])

    view.zoom_fit()
    cell = view.active_cellview().cell

    view_box = cell.dbbox()
    view_box.left -= margin
    view_box.bottom -= margin
    view_box.right += margin
    view_box.top += margin

    x_incr = int(view_box.width() / xbins)
    y_incr = int(view_box.height() / ybins)

    if (view_box.width() > view_box.height()):
        y_px = vertical_resolution
        x_px = int((float(x_incr) / y_incr) * y_px)
    else:
        x_px = horizontal_resolution
        y_px = int((float(y_incr) / x_incr) * x_px)

    for x in range(xbins):
        for y in range(ybins):
            yidx = ybins - y - 1
            output_file = f"{design}_X{x}_Y{yidx}.png"

            x_start = view_box.left + x_incr * x
            y_start = view_box.bottom + y_incr * y

            subbox = pya.DBox(
                x_start,
                y_start,
                x_start + x_incr,
                y_start + y_incr)

            sub_img_spec = {
                "width": x_px,
                "height": y_px,
                "linewidth": linewidth,
                "oversampling": oversampling,
                "resolution": 0,
                "target_box": subbox,
                "monochrome": False
            }

            img_path = os.path.join('outputs', output_file)
            print(f'[INFO] Saving screenshot (({subbox.left}, {subbox.bottom}), '
                  f'({subbox.right}, {subbox.top})) to {img_path}')
            view.save_image_with_options(
                img_path,
                sub_img_spec["width"],
                sub_img_spec["height"],
                sub_img_spec["linewidth"],
                sub_img_spec["oversampling"],
                sub_img_spec["resolution"],
                sub_img_spec["target_box"],
                sub_img_spec["monochrome"]
            )


def main():
    # SC_ROOT provided by CLI, and is only accessible when this is main module
    sys.path.append(SC_ROOT)  # noqa: F821

    from tools.klayout.klayout_utils import (
        technology,
        get_schema
    )

    schema = get_schema(manifest='sc_manifest.json')

    flow = schema.get('option', 'flow')
    step = schema.get('arg', 'step')
    index = schema.get('arg', 'index')
    task = schema.get('flowgraph', flow, step, index, 'task')

    design = schema.get('option', 'entrypoint')
    if not design:
        design = schema.get('design')

    if 'show_filepath' in schema.getkeys('tool', 'klayout', 'task', task, 'var') and \
       schema.get('tool', 'klayout', 'task', task, 'var', 'show_filepath',
                  step=step, index=index):
        sc_filename = schema.get('tool', 'klayout', 'task', task, 'var', 'show_filepath',
                                 step=step, index=index)[0]
    else:
        sc_fileext = schema.get('tool', 'klayout', 'task', task, 'var', 'show_filetype',
                                step=step, index=index)[0]
        for ext in (f'{sc_fileext}.gz', sc_fileext):
            sc_filename = f"inputs/{design}.{ext}"
            if os.path.exists(sc_filename):
                break

    sc_exit = schema.get('tool', 'klayout', 'task', task, 'var', 'show_exit',
                         step=step, index=index) == ["true"]

    show(schema, technology(design, schema), sc_filename, f'outputs/{design}.png',
         screenshot=(task == 'screenshot'))

    if sc_exit:
        pya.Application.instance().exit(0)


if __name__ == '__main__':
    main()



import siliconcompiler

from siliconcompiler.flows._common import _make_docs
from siliconcompiler.tools.verilator import lint as verilator_lint
from siliconcompiler.tools.slang import lint as slang_lint


###########################################################################
# Flowgraph Setup
############################################################################
def setup(tool='verilator'):
    '''
    An RTL linting flow.
    '''

    flowname = 'lintflow'
    flow = siliconcompiler.Flow(flowname)

    if tool == 'verilator':
        flow.node(flowname, 'lint', verilator_lint)
    elif tool == 'slang':
        flow.node(flowname, 'lint', slang_lint)
    else:
        raise ValueError(f'Unsupported lint tool: {tool}')

    return flow


##################################################
if __name__ == "__main__":
    chip = siliconcompiler.Chip('design')
    _make_docs(chip)
    flow = setup()
    chip.use(flow)
    chip.write_flowgraph(f"{flow.top()}.png", flow=flow.top())



from siliconcompiler.report import report
from siliconcompiler.flowgraph import _get_flowgraph_exit_nodes, _get_flowgraph_entry_nodes
from siliconcompiler.tools._common import get_tool_task
from siliconcompiler import NodeStatus

from streamlit_agraph import Node, Edge, Config


# for flowgraph
NODE_COLORS = {
    NodeStatus.SUCCESS: '#70db70',  # green

    NodeStatus.SKIPPED: '#ffc299',  # orange

    NodeStatus.PENDING: '#6699ff',  # blue
    NodeStatus.QUEUED: '#6699ff',  # blue

    NodeStatus.RUNNING: '#ffff4d',  # yellow

    NodeStatus.ERROR: '#ff1a1a',  # red
    NodeStatus.TIMEOUT: '#ff1a1a',  # red

    "Unknown": '#6699ff',  # blue
}


def get_nodes_and_edges(chip):
    """
    Returns the nodes and edges required to make a streamlit_agraph.

    Args:
        chip (Chip) : The chip object that contains the schema read from.
    """
    nodes = []
    edges = []

    default_node_border_width = 1
    successful_path_node_width = 3
    default_edge_width = 3
    successful_path_edge_width = 5

    node_dependencies = report.get_flowgraph_edges(chip)
    successful_path = report.get_flowgraph_path(chip)

    entry_exit_nodes = _get_flowgraph_entry_nodes(chip, chip.get('option', 'flow')) + \
        _get_flowgraph_exit_nodes(chip, chip.get('option', 'flow'))

    for step, index in node_dependencies:
        # Build node
        node_border_width = default_node_border_width
        if (step, index) in entry_exit_nodes:
            node_border_width = successful_path_node_width

        node_status = chip.get('record', 'status', step=step, index=index)
        if node_status not in NODE_COLORS:
            node_status = "Unknown"
        node_color = NODE_COLORS[node_status]

        tool, task = get_tool_task(chip, step, index)
        node_name = f'{step}{index}'
        label = node_name + "\n" + tool + "/" + task
        if tool == 'builtin':
            label = node_name + "\n" + tool

        nodes.append(Node(
            id=node_name,
            label=label,
            color=node_color,
            opacity=1,
            borderWidth=node_border_width,
            shape='oval',
            fixed=True))

        # Build edges
        path_taken = chip.get('record', 'inputnode', step=step, index=index)
        all_edges = set([*node_dependencies[step, index], *path_taken])
        for source_step, source_index in all_edges:
            edge_width = default_edge_width
            if (step, index) in successful_path and \
               (source_step, source_index) in successful_path:
                edge_width = successful_path_edge_width

            dashes = False
            color = 'black'
            if (source_step, source_index) not in path_taken:
                color = 'gray'
                dashes = True
            elif node_status != NodeStatus.SUCCESS:
                color = 'gray'
            elif NodeStatus.is_waiting(node_status) or NodeStatus.is_running(node_status):
                color = 'blue'
                dashes = True

            edges.append(Edge(
                source=f'{source_step}{source_index}',
                target=node_name,
                dir='up',
                width=edge_width,
                color=color,
                dashes=dashes))

    return nodes, edges


def get_graph_config():
    return Config(
        width='100%',
        height=1000,
        directed=True,
        physics=False,
        hierarchical=True,
        nodeSpacing=150,
        levelSeparation=100,
        sortMethod='directed')
